"""
Copyright (c) 2008-2023 synodriver <diguohuangjiajinweijun@gmail.com>
"""
import os
import sys

from cffi import FFI

define_macros = []
if sys.platform.startswith("win"):
    extra_compile_args = ["-openmp:experimental"]
    extra_link_args = ["-openmp:experimental"]
    define_macros.append(("LIBQDIFF_PUBLIC_API", "__declspec(dllexport)"))
elif sys.platform.startswith("darwin"):
    os.system("brew install libomp")
    extra_compile_args = ["-Xpreprocessor", "-fopenmp"]
    extra_link_args = ["-L/usr/local/lib", "-lomp"]
else:
    extra_compile_args = ["-fopenmp"]
    extra_link_args = ["-fopenmp"]


ffibuilder = FFI()
ffibuilder.cdef(
    """
#define QBERR_OK 0
#define QBERR_NOMEM 1
#define QBERR_IOERR 2
#define QBERR_TRUNCPATCH 3
#define QBERR_BADPATCH 4
#define QBERR_BADCKSUM 5
#define QBERR_LZMAERR 6
#define QBERR_SAIS 7

int qbdiff_compute(const uint8_t * old, const uint8_t * new, size_t old_len, size_t new_len,
                                       FILE * diff_file);
int qbdiff_patch(const uint8_t * old, const uint8_t * patch, size_t old_len, size_t patch_len,
                                     FILE * new_file);
const char * qbdiff_version(void);
const char * qbdiff_error(int code);

FILE *fopen   (const char *filename, const char  *opentype);
int    fflush (FILE *stream);
int  fclose   (FILE *stream);
    """
)

source = """
#include "libqbdiff.h"
"""
c_sources = [
    "./dep/src/blake2b.c",
    "./dep/src/libqbdiff.c",
    "./dep/src/libsais.c",
    "./dep/src/libsais64.c",
]

ffibuilder.set_source(
    "qbdiff.backends.cffi._qbdiff",
    source,
    sources=c_sources,
    include_dirs=["./dep/include", os.getenv("INCLUDE", "")],
    define_macros=define_macros,
    extra_compile_args=extra_compile_args,
    extra_link_args=extra_link_args,
    extra_objects=[os.getenv("LIB", "")],
)

if __name__ == "__main__":
    ffibuilder.compile()



"""
Copyright (c) 2008-2023 synodriver <diguohuangjiajinweijun@gmail.com>
"""
from qbdiff.backends.cffi._qbdiff import ffi, lib

QBERR_OK           = lib.QBERR_OK
QBERR_NOMEM        = lib.QBERR_NOMEM
QBERR_IOERR        = lib.QBERR_IOERR
QBERR_TRUNCPATCH   = lib.QBERR_TRUNCPATCH
QBERR_BADPATCH     = lib.QBERR_BADPATCH
QBERR_BADCKSUM     = lib.QBERR_BADCKSUM
QBERR_LZMAERR      = lib.QBERR_LZMAERR
QBERR_SAIS         = lib.QBERR_SAIS

def version() -> str:
    return ffi.string(lib.qbdiff_version()).decode()


def error(code: int) -> str:
    return ffi.string(lib.qbdiff_error(code)).decode()


def compute(old: bytes, new_: bytes, diff_file: str) -> int:
    # cdef int ret
    f = lib.fopen(diff_file.encode(), b"wb")
    if f == ffi.NULL:
        raise ValueError("can not open file")
    try:
        ret = lib.qbdiff_compute(
            ffi.from_buffer(old), ffi.from_buffer(new_), len(old), len(new_), f
        )
        if ret != lib.QBERR_OK:
            raise ValueError(
                f"Failed to create delta (error {ret}: {ffi.string(lib.qbdiff_error(ret)).decode()}"
            )
        return ret
    finally:
        lib.fflush(f)
        lib.fclose(f)


def patch(old: bytes, patch_: bytes, new_file: str) -> int:
    f = lib.fopen(new_file.encode(), b"wb")
    if f == ffi.NULL:
        raise ValueError("can not open file")
    try:
        ret = lib.qbdiff_patch(
            ffi.from_buffer(old), ffi.from_buffer(patch_), len(old), len(patch_), f
        )
        if ret != lib.QBERR_OK:
            raise ValueError(
                f"Failed to patch (error {ret}: {ffi.string(lib.qbdiff_error(ret)).decode()}"
            )
        return ret
    finally:
        lib.fflush(f)
        lib.fclose(f)



import os
import platform

for f in os.listdir("dist"):
    if "linux" in f:
        os.rename(
            os.path.join("dist", f),
            os.path.join("dist", f.replace("linux", "manylinux2014")),
        )
    elif "macosx" in f:
        if platform.machine() == "x86_64":
            os.rename(
                os.path.join("dist", f),
                os.path.join("dist", f.replace("universal2", "x86_64")),
            )
        else:
            os.rename(
                os.path.join("dist", f),
                os.path.join("dist", f.replace("universal2", "arm64")),
            )



# cython: language_level=3
# cython: cdivision=True
from libc.stdint cimport uint8_t
from libc.stdio cimport FILE


cdef extern from "libqbdiff.h" nogil:
    int QBERR_OK
    int QBERR_NOMEM
    int QBERR_IOERR
    int QBERR_TRUNCPATCH
    int QBERR_BADPATCH
    int QBERR_BADCKSUM
    int QBERR_LZMAERR
    int QBERR_SAIS
    int qbdiff_compute(const uint8_t * old, const uint8_t * new_, size_t old_len, size_t new_len,
                       FILE * diff_file)
    int qbdiff_patch(const uint8_t * old, const uint8_t * patch, size_t old_len, size_t patch_len,
                     FILE * new_file)
    const char * qbdiff_version()
    const char * qbdiff_error(int code)


"""
Copyright (c) 2008-2023 synodriver <diguohuangjiajinweijun@gmail.com>
"""
from qbdiff.backends.cython._qbdiff import (
    QBERR_BADCKSUM,
    QBERR_BADPATCH,
    QBERR_IOERR,
    QBERR_LZMAERR,
    QBERR_NOMEM,
    QBERR_OK,
    QBERR_SAIS,
    QBERR_TRUNCPATCH,
    compute,
    error,
    patch,
    version,
)



QBERR_BADCKSUM: int
QBERR_BADPATCH: int
QBERR_IOERR: int
QBERR_LZMAERR: int
QBERR_NOMEM: int
QBERR_OK: int
QBERR_SAIS: int
QBERR_TRUNCPATCH: int

def version() -> str: ...
def error(code: int) -> str: ...
def compute(old: bytes, new_: bytes, diff_file: str) -> int: ...
def patch(old: bytes, patch_: bytes, new_file: str) -> int: ...



# -*- coding: utf-8 -*-
import glob
import os
import re
import sys
from collections import defaultdict

try:
    from Cython.Build import cythonize
except ImportError:
    Cython = None
from setuptools import Extension, find_packages, setup
from setuptools.command.build_ext import build_ext

BUILD_ARGS = defaultdict(lambda: ["-O3", "-g0"])

for compiler, args in [
    ("msvc", ["/EHsc", "/DHUNSPELL_STATIC", "/Oi", "/O2", "/Ot"]),
    ("gcc", ["-O3", "-g0"]),
]:
    BUILD_ARGS[compiler] = args

define_macros = []
if sys.platform.startswith("win"):
    extra_compile_args = ["-openmp:experimental"]
    extra_link_args = ["-openmp:experimental"]
    define_macros.append(("LIBQDIFF_PUBLIC_API", "__declspec(dllexport)"))
elif sys.platform.startswith("darwin"):
    os.system("brew install libomp")
    extra_compile_args = ["-Xpreprocessor", "-fopenmp"]
    extra_link_args = ["-L/usr/local/lib", "-lomp"]
else:
    extra_compile_args = ["-fopenmp"]
    extra_link_args = ["-fopenmp"]


def has_option(name: str) -> bool:
    if name in sys.argv[1:]:
        sys.argv.remove(name)
        return True
    return False


class build_ext_compiler_check(build_ext):
    def build_extensions(self):
        compiler = self.compiler.compiler_type
        args = BUILD_ARGS[compiler]
        for ext in self.extensions:
            ext.extra_compile_args.extend(args)
            # if self.compiler.compiler_type == "msvc":
            #     ext.define_macros.extend([("restrict", "__restrict")])
        super().build_extensions()


c_sources = ["qbdiff/backends/cython/_qbdiff.pyx"] + [
    "./dep/src/blake2b.c",
    "./dep/src/libqbdiff.c",
    "./dep/src/libsais.c",
    "./dep/src/libsais64.c",
]

extensions = [
    Extension(
        "qbdiff.backends.cython._qbdiff",
        c_sources,
        include_dirs=["./dep/include", os.getenv("INCLUDE", "")],
        define_macros=define_macros,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        extra_objects=[os.getenv("LIB", "")],
    ),
]
cffi_modules = ["qbdiff/backends/cffi/build.py:ffibuilder"]


def get_dis():
    with open("README.markdown", "r", encoding="utf-8") as f:
        return f.read()


def get_version() -> str:
    path = os.path.join(
        os.path.abspath(os.path.dirname(__file__)), "qbdiff", "__init__.py"
    )
    with open(path, "r", encoding="utf-8") as f:
        data = f.read()
    result = re.findall(r"(?<=__version__ = \")\S+(?=\")", data)
    return result[0]


packages = find_packages(exclude=("test", "tests.*", "test*"))

setup_requires = []
install_requires = []
setup_kw = {}
if has_option("--use-cython"):
    print("building cython")
    setup_requires.append("cython")
    setup_kw["ext_modules"] = cythonize(
        extensions,
        compiler_directives={
            "cdivision": True,
            "embedsignature": True,
            "boundscheck": False,
            "wraparound": False,
        },
    )
if has_option("--use-cffi"):
    print("building cffi")
    setup_requires.append("cffi>=1.0.0")
    install_requires.append("cffi>=1.0.0")
    setup_kw["cffi_modules"] = cffi_modules


def main():
    version: str = get_version()

    dis = get_dis()
    setup(
        name="qbdiff",
        version=version,
        url="https://github.com/synodriver/pyqbdiff",
        packages=packages,
        keywords=["qbdiff", "delta", "path"],
        description="python binding for qbdiff",
        long_description_content_type="text/markdown",
        long_description=dis,
        author="synodriver",
        author_email="diguohuangjiajinweijun@gmail.com",
        python_requires=">=3.6",
        setup_requires=setup_requires,
        install_requires=install_requires,
        license="BSD",
        classifiers=[
            "Development Status :: 4 - Beta",
            "Operating System :: OS Independent",
            "Topic :: System :: Archiving",
            "Topic :: System :: Archiving :: Compression",
            "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
            "Programming Language :: C",
            "Programming Language :: Cython",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
            "Programming Language :: Python :: 3.12",
            "Programming Language :: Python :: Implementation :: CPython",
            "Programming Language :: Python :: Implementation :: PyPy",
        ],
        include_package_data=True,
        zip_safe=False,
        cmdclass={"build_ext": build_ext_compiler_check},
        **setup_kw
    )


if __name__ == "__main__":
    main()



"""
Copyright (c) 2008-2023 synodriver <diguohuangjiajinweijun@gmail.com>
"""
from unittest import TestCase

from qbdiff import compute, error, patch, version


class Testcy(TestCase):
    def setUp(self) -> None:
        pass

    def tearDown(self) -> None:
        pass

    def test_version(self):
        print(version())
        self.assertTrue(isinstance(version(), str))

    def test_error(self):
        print(error(1))
        self.assertTrue(isinstance(error(1), str))

    def test_compute(self):
        old = b"1234"
        new_ = b"12345"
        compute(old, new_, "diff_tmp.bin")
        with open("diff_tmp.bin", "rb") as f:
            diff = f.read()
        patch(old, diff, "new.bin")
        with open("new.bin", "rb") as f:
            newf = f.read()
        self.assertEquals(new_, newf, "patch fail")


if __name__ == "__main__":
    import unittest

    unittest.main()



# cython: language_level=3
# cython: cdivision=True
from libc.stdint cimport uint8_t
from libc.stdio cimport FILE, fclose, fflush, fopen

from qbdiff.backends.cython cimport qbdiff
from qbdiff.backends.cython.qbdiff cimport (qbdiff_compute, qbdiff_error,
                                            qbdiff_patch, qbdiff_version)

QBERR_OK           = qbdiff.QBERR_OK
QBERR_NOMEM        = qbdiff.QBERR_NOMEM
QBERR_IOERR        = qbdiff.QBERR_IOERR
QBERR_TRUNCPATCH   = qbdiff.QBERR_TRUNCPATCH
QBERR_BADPATCH     = qbdiff.QBERR_BADPATCH
QBERR_BADCKSUM     = qbdiff.QBERR_BADCKSUM
QBERR_LZMAERR      = qbdiff.QBERR_LZMAERR
QBERR_SAIS         = qbdiff.QBERR_SAIS


cdef extern from "Python.h":
    const char *PyUnicode_AsUTF8(object u)

cpdef inline str version():
    return (<bytes>qbdiff_version()).decode()

cpdef inline str error(int code):
    return (<bytes>qbdiff_error(code)).decode()

cpdef inline int compute(const uint8_t[::1] old, const uint8_t[::1] new_, str diff_file) except -1:
    cdef int ret
    cdef FILE* f = fopen(PyUnicode_AsUTF8(diff_file),"wb")
    if f==NULL:
        raise ValueError("can not open file")
    try:
        with nogil:
            ret = qbdiff_compute(&old[0],&new_[0],<size_t>old.shape[0], <size_t>new_.shape[0], f)
        if ret != qbdiff.QBERR_OK:
            raise ValueError(f"Failed to create delta (error {ret}: {(<bytes>qbdiff_error(ret)).decode()}")
        return ret
    finally:
        fflush(f)
        fclose(f)

cpdef inline int patch(const uint8_t[::1] old, const uint8_t[::1] patch_, str new_file) except -1:
    cdef int ret
    cdef FILE* f = fopen(PyUnicode_AsUTF8(new_file),"wb")
    if f==NULL:
        raise ValueError("can not open file")
    try:
        with nogil:
            ret = qbdiff_patch(&old[0],&patch_[0],<size_t>old.shape[0], <size_t>patch_.shape[0], f)
        if ret != qbdiff.QBERR_OK:
            raise ValueError(f"Failed to patch (error {ret}: {(<bytes>qbdiff_error(ret)).decode()}")
        return ret
    finally:
        fflush(f)
        fclose(f)
